open Exp
open Bigarray
open Gc

type env =  {params: bool array; vars: bool array}
type lenv = {lparams: int array; lvars: int array}

type state = Carries of int array | Trash | Maybe

let int_type = Bigarray.int16_unsigned
type int_type_elt = Bigarray.int16_unsigned_elt

let rec pow3 = function
  | 0 -> 1
  | i -> 3 * (pow3 (i-1))

let bitrev w z =
  let rec f x accu = function
    | 0 -> accu
    | i -> f (x/2) (2*accu+(x mod 2)) (i-1)
  in f z 0 w

let eval s (Carries cr) e =
  let ncarries = Array.copy cr in
  let cache = Array.make (Array.length cr) None in
  let rec ev = function
  | Const c     -> c
  | Param i     -> e.params.(i)
  | Var   i     -> e.vars.(i)
  (* Lazy evaluation killed the radiostar :-( *)
  | Bool (f,a,b)   -> let u = ev a and v = ev b in f u v
  | HW (i,w,a) -> let x = cr.(i)+(if ev a then 1 else 0) in
                  ncarries.(i) <- x;
                  x<=w;
  | Sum (i,a,b) ->
      (match cache.(i) with
	 |None -> let b =
	     (match cr.(i)+(if ev a then 1 else 0)+(if ev b then 1 else 0) with
		| 0 -> ncarries.(i) <- 0; false
		| 1 -> ncarries.(i) <- 0; true
		| 2 -> ncarries.(i) <- 1; false
		| 3 -> ncarries.(i) <- 1; true)
	   in cache.(i) <- Some b; b
	 | Some b -> b)
  | Sum3 (i,a,b,c) ->
      (match cache.(i) with
	 |None -> let b =
	     (match cr.(i)+(if ev a then 1 else 0)
		+(if ev b then 1 else 0)+(if ev c then 1 else 0) with
		  | 0 -> ncarries.(i) <- 0; false
		  | 1 -> ncarries.(i) <- 0; true
		  | 2 -> ncarries.(i) <- 1; false
		  | 3 -> ncarries.(i) <- 1; true
		  | 4 -> ncarries.(i) <- 2; false
		  | 5 -> ncarries.(i) <- 2; true)
	   in cache.(i) <- Some b; b
	 | Some b -> b)
  | Mux (i, l)        ->                (* Warning: Mux is Lazy! *)
      let rec f accu = function
        | (0, _)    -> accu
        | (x, h::t) -> f (2*accu + if ev h then 1 else 0) (x-1, t)
      in let idx = f 0 (i,l) in
      ev (List.nth l (i+idx))
  in
  if List.fold_left (&&) true (List.map ev s.exps) then Carries ncarries else Trash

let rec lazy_bool f u v =
  let n x = if x then 1 else 0 in
  let b = function
    | 0 -> false
    | 1 -> true
    | _ -> failwith "ICE"
  in match u with
    | -1 -> if lazy_bool f 0 v = lazy_bool f 1 v then lazy_bool f 0 v else -1
    |  _ -> match v with
        | -1 -> if f (b u) false = f (b u) true then n (f (b u) false) else -1
        |  _ -> n (f (b u) (b v))
            
let eval_lazy s (Carries cr) e =
  let ncarries = Array.copy cr in
  let cache = Array.make (Array.length cr) None in
  let rec ev = function
  | Const c     -> if c then 1 else 0
  | Param i     -> e.lparams.(i)
  | Var   i     -> e.lvars.(i)
  (* Lazy evaluation killed the radiostar :-( *)
  | Bool (f,a,b)   -> let u = ev a and v = ev b in lazy_bool f u v
  | Sum (i,a,b) ->
    (match cache.(i) with
      |None -> let b =
                 if cr.(i) == -1 || ev a == -1 || ev b == -1 then
                   -1
                 else
	           (match cr.(i)+(ev a)+(ev b) with
		     | 0 -> ncarries.(i) <- 0; 0
		     | 1 -> ncarries.(i) <- 0; 1
		     | 2 -> ncarries.(i) <- 1; 0
		     | 3 -> ncarries.(i) <- 1; 1)
	       in cache.(i) <- Some b; b
      | Some b -> b)
  | Sum3 (i,a,b,c) ->
    (match cache.(i) with
      |None -> let b =
                 if cr.(i) == -1 || ev a == -1 || ev b == -1 || ev c == -1 then
                   -1
                 else
	           (match cr.(i)+(ev a)+(ev b)+(ev c) with
	             | 0 -> ncarries.(i) <- 0; 0
		     | 1 -> ncarries.(i) <- 0; 1
		     | 2 -> ncarries.(i) <- 1; 0
		     | 3 -> ncarries.(i) <- 1; 1
		     | 4 -> ncarries.(i) <- 2; 0
		     | 5 -> ncarries.(i) <- 2; 1)
	       in cache.(i) <- Some b; b
      | Some b -> b)
  | Mux (i, l)        ->                (* Warning: Mux is Lazy! *)
    let rec f accu = function
      | (0, _)    -> accu
      | (x, h::t) -> let z = ev h in
                     if z == -1 then -1 else f (2*accu + z) (x-1, t)
    in let idx = f 0 (i,l) in
       ev (List.nth l (i+idx))
  in
  let e = List.fold_left (fun x y -> Bool((&&), x, y))
    (List.hd s.exps) (List.tl s.exps) in
  match ev e with
    | 1  -> Carries ncarries
    | 0  -> Trash
    | -1 -> Maybe

let verbose = ref false

let debug s =
  if !verbose then (
    output_string stderr s;
    flush_all ()
  )

exception Over

let rec next_tab t i =
  if t.(i) then begin
    t.(i) <- false;
    next_tab t (i+1)
    end
  else
    t.(i) <- true

let next_env e =
  try next_tab e.params 0
  with Invalid_argument "index out of bounds" ->
    debug ".";
    try next_tab e.vars 0
    with Invalid_argument "index out of bounds" -> raise Over

let copy_env e =
  {params= Array.copy e.params; vars= Array.copy e.vars}

let env_to_int e =
  Array.fold_left (fun x b -> 2*x + if b then 1 else 0) 
    (Array.fold_left (fun x b -> 2*x + if b then 1 else 0) 0 e.params)
    e.vars

let lenv_to_sparse e j =
  let k = (Array.length e.lparams)/j in 
  let t = Array.make k 0 in
  for a=0 to k-1 do
    for b=0 to j-1 do
      if e.lparams.(a*j+b) == -1 then
        failwith "ICE";
      t.(a) <- 2*t.(a) + e.lparams.(a*j+b)
    done;
  done;
  t
    
let int_to_env syst u =
  let f x m i =
    (m land (x lsl i)) <> 0 in
  {
    params= Array.init syst.nbparam (f (u lsr syst.nbvar) (1 lsl (syst.nbparam-1)));
    vars  = Array.init syst.nbvar   (f u (1 lsl (syst.nbvar-1)))
  }

(* type transitions = int array array *)
type 'a transitions = (int, 'a, Bigarray.c_layout)
    Bigarray.Array2.t

type 'a map_array = {fd: Unix.file_descr;
                     mutable ba_large: 'a transitions;
                     mutable ba: 'a transitions}
    
let create_map_array_sub s l kind =
  let fd = Unix.openfile s [ Unix.O_RDWR; Unix.O_CREAT; Unix.O_TRUNC ] 0o640 in
  let ba_large = Array2.map_file fd kind Bigarray.c_layout true 16 l in
  let ba = Array2.sub_left ba_large 0 0 in
  {fd = fd; ba_large = ba_large; ba = ba}

let close_map_array ma =                (* Does not really work *)
  ma.ba_large <- Array2.create (Array2.kind ma.ba) Bigarray.c_layout 0 0;
  ma.ba <- ma.ba_large;
  Unix.close ma.fd

let create_map_array s l kind =
  let ma = create_map_array_sub s l kind in
  Gc.finalise close_map_array ma;
  ma

let create_map_array_anon l kind =
  let tmpf = Filename.temp_file "BigArray" ".mmap" in
  let ma = create_map_array tmpf l kind in
  Sys.remove tmpf;
  ma

let resize_map_array ma n =
  let l = Array2.dim2 ma.ba in
  let k = Array2.dim1 ma.ba_large in
  let nn = if n < 2*k then 2*k else n in
  if (n > k) then 
    ma.ba_large <-
      Array2.map_file ma.fd (Array2.kind ma.ba) Bigarray.c_layout
      true nn l;
  ma.ba <- Array2.sub_left ma.ba_large 0 n

let reset_map_array ma n p =
  ma.ba_large <-
    Array2.map_file ma.fd (Array2.kind ma.ba) Bigarray.c_layout
    true n p;
  ma.ba <- ma.ba_large
    
let grow_map_array ma =
  let n = Array2.dim1 ma.ba in
  resize_map_array ma (n+1)

 
let msign x =
  if x == 65535 then -1 else x

let build syst =
  let automata = create_map_array_anon
    (1 lsl (syst.nbparam + syst.nbvar)) int_type in
  let idx      = Hashtbl.create 64 in
  Hashtbl.add idx Trash (-1);
  let nb       = ref 0 in
  let z = Carries (Array.copy syst.initcarry) in
  let rec trans s =
    grow_map_array automata;
    let t = Array2.slice_left automata.ba ((Array2.dim1 automata.ba)-1) in
    Array1.fill t (-1);
    Hashtbl.add idx s !nb;
    (* Printf.printf "\nState %i: " (!nb); *)
    (* match s with *)
    (* Carries x -> Array.iter (Printf.printf "%i") x; *)
    incr nb;
    let e = {params= Array.make syst.nbparam false; vars= Array.make syst.nbvar false} in
    let rec loop () =
      let _ = match eval syst s e with
        | Trash           -> ()
        | Carries x as ns -> if not (Hashtbl.mem idx ns) then trans ns;
          t.{env_to_int e} <- Hashtbl.find idx ns; in
      next_env e; loop ()
    in
    try
      loop ()
    with Over -> debug "\n";
  in trans z;
  automata.ba

external output_mask_c : 
  Unix.file_descr -> int -> int -> unit
    = "output_mask_c"

external output_int_c : 
  Unix.file_descr -> int -> unit
    = "output_int_c"

let rec next_tab_lazy t i skip =
  if skip then (* No need to recurse over unknowns *)
    match t.(i) with
      | -1 -> next_tab_lazy t (i+1) true
      |  _ -> next_tab_lazy t i false
  else
    match t.(i) with
      |  0 -> t.(i) <- 1;
      |  1 -> t.(i) <- -1; next_tab_lazy t (i+1) skip
      | -1 -> if i+1 < Array.length t && t.(i+1) == -1 then next_tab_lazy t (i+1) skip else t.(i) <- 0
          
let next_env_lazy e skip =
  try next_tab_lazy e.lvars 0 skip
  with Invalid_argument "index out of bounds" ->
    debug ".";
    try next_tab_lazy e.lparams 0 skip
    with Invalid_argument "index out of bounds" -> raise Over

let build_sparse syst fd bits =
  let automata = ref [] in
  let idx      = Hashtbl.create 64 in
  let sets     = Hashtbl.create 64 in
  Hashtbl.add idx Trash (-1);
  let z = Carries (Array.copy syst.initcarry) in
  Hashtbl.add idx z 0;
  Hashtbl.add sets 0 z;
  let rec trans cur max =
    debug (Printf.sprintf "\r%i/%i" cur max);
    if cur <= max then
      let line = ref [] in
      let s = Hashtbl.find sets cur in
      let max2 = ref max in
      let te = Array.make (syst.nbparam+syst.nbvar) (-1) in
      let rec loop () =
        let e = {lparams= Array.sub te 0 syst.nbparam;
                 lvars= Array.sub te syst.nbparam syst.nbvar} in

        let b = match eval_lazy syst s e with
          | Trash           -> true     (* Skip unknowns *)
          | Maybe           -> false    (* Recurse over unknowns *)
          | Carries x as ns -> if not (Hashtbl.mem idx ns) then
              begin
                incr max2;
                Hashtbl.add idx ns !max2;
                Hashtbl.add sets !max2 ns;
              end;
            let t = lenv_to_sparse e bits in
            (* Printf.printf "%i -> %i: " cur (Hashtbl.find idx ns); *)
            (* Array.iter (fun x-> Printf.printf "%i, " (31-x)) t; *)
            (* Printf.printf "\n"; *)
            flush_all ();
            line := (t,(Hashtbl.find idx ns))::!line;
            false
        in
        (* for i=0 to Array.length e.lparams-1 do *)
        (*   Printf.printf "%2i " e.lparams.(i); *)
        (* done; *)
        (* for i=0 to Array.length e.lvars-1 do *)
        (*   Printf.printf "%2i " e.lvars.(i); *)
        (* done; *)
        (* Printf.printf "-> %b\n" b; *)

        next_tab_lazy te 0 b; loop ()
      in begin
        try
          loop ()
        with Invalid_argument "index out of bounds" -> ();
      end;
      automata := !line::!automata;
      trans (cur+1) !max2
  in trans 0 0;
  debug "\n";
(* Verify that the automata is suitable *)
  let n = List.length (List.hd !automata) in
  List.iter (fun x-> if List.length x <> n then failwith "Bad system: sparse automata can't be used") !automata;
  let rec dup_p = function
    |[]|[_] -> ()
    |(_,h)::t   -> if List.exists (fun (_,x) -> x = h) t then failwith "Bad system: sparse automata can't be used"
  in List.iter dup_p !automata;
  Printf.printf "Seems good\n";
  output_int_c fd (List.length !automata);
  output_int_c fd (syst.nbparam/bits);
  output_int_c fd n;
  List.iter (
    List.iter (fun (t,x) -> Array.iter (output_mask_c fd bits) t; output_int_c fd x)
  ) (List.rev !automata)

  
module type AutomataIface =  sig
  type v
  val v_to_int : v -> int
  val build : syst -> int_type_elt transitions
  val sizealpha : syst -> int
end


module Int = struct
  type t = int
  let compare = Pervasives.compare
end

module IntSet = Set.Make(Int)

module SetSet = Set.Make(IntSet)


  let minimize at =
    let sz = Array2.dim2 at in
    debug "Minimizing";
    (* Moore's minimization algorithm *)

    (* Intersection of partitions using radix sort *)
    (* Order is broken, but partitions should be OK *)
    let reorder p =
      let map = Array.make (Array.length p) (-1) in
      let n = ref 0 in
      let f x =
        if map.(x) == -1 then (map.(x) <- !n; incr n);
        map.(x)
      in Array.map f p;
    in

    let ma = create_map_array_anon (Array2.dim1 at) int_type in
    resize_map_array ma (sz+1);

    let refine p = 
      debug ".";
      let n = Array.length p in
      let m = Array.fold_left max 0 p in
      if m+1 = n then p else
        (* let l = Array2.create Bigarray.int Bigarray.c_layout ((sz)+1) n in *)
        let l = ma.ba in
        let read i j=
          (* debug (Printf.sprintf "[%i,%i] (%i,%i)\n" i j (Array2.dim1 at) (Array2.dim2 at)); *)
          if j = (Array2.dim2 at)
          then p.(i)
          else
            let y = msign at.{i,j} in
            if y <> -1 then
              (p.(y)+1)
            else 0               (* 0 is for unreached *)
        in
        let cnt = ref 0 in
        let rec transpose i di j dj =
          if di <= 16 && dj <= 16 then
            begin
              for u=i to i+di-1 do
                for v=j to j+dj-1 do
                  l.{v,u} <- read u v
                done
              done;
              incr cnt;
              if !cnt mod (((Array2.dim1 at)*(Array2.dim2 at)/25600)+1) = 0 then
                debug "*";
            end
          else if di < dj then
            let t = (dj/2) in
            transpose i di j t;
            transpose i di (j+t) (dj-t)
          else
            let t = (di/2) in
            transpose i t j dj;
            transpose (i+t) (di-t) j dj
        in transpose 0 (Array2.dim1 at) 0 (Array2.dim2 at);
        Array1.blit (Array1.of_array int_type Bigarray.c_layout p)
          (Array2.slice_left l (sz));
        debug "#";
        
        let inter =
          let rec pass b i =
            if i < 0 then b else
              let buckets = Array.make n [] in
              let f x =
                buckets.(l.{i,x}) <- x::buckets.(l.{i,x});
              in Array.iter (List.iter f) b;
              pass buckets (i-1);
          in let buckets = pass (Array.init n (fun i -> [i])) ((Array2.dim1 l)-1) in
          debug "/";
          let samep x y =
            let rec samei i =
              if (i<0) then true else
                read x i = read y i && (samei (i-1))
            in
            (y <> -1) && samei ((Array2.dim1 l)-1)
          in let partition = Array.make n (-1) in
          let xx = ref (-1) in
          let p = ref (-1) in
          let f x =
            if (samep x !xx) then () else incr p;
            partition.(x) <- !p;
            xx := x;
          in Array.iter (List.iter f) buckets;
          debug (Printf.sprintf "%i/" (!p+1));
          partition
        in
        reorder inter
    in
    
    let rec fixedpoint f x =
      let y = f x in
      if x = y then
        x
      else
        fixedpoint f y
    in
   
    let p = fixedpoint refine (Array.make (Array2.dim1 at) 0) in
    let partition = p in
    let m = Array.fold_left max 0 partition in
    let choose = Array.make (m+1) (-1) in
    Array.iteri (fun i x -> choose.(x) <- i) partition;

    (* Output automata *)

    let automata = Array2.create int_type Bigarray.c_layout
      (m+1) sz in
      (* Array.make_matrix (m+1) (sz) (-1) in *)
    for i = 0 to m do
      for j = 0 to sz-1 do
        let y = msign at.{choose.(i),j} in
        if y <> -1 then
          automata.{i,j} <- partition.(y)
        else
          automata.{i,j} <- -1
      done;
    done;

    debug (Printf.sprintf "\nMinimizing done %i -> %i\n"
             (Array2.dim1 at) (m+1));
    if partition.(0) <> 0 then
      failwith "ICE";
    automata


module Minimizer(X: AutomataIface): AutomataIface = struct
  type v = X.v
  let sizealpha = X.sizealpha
  let v_to_int = X.v_to_int
  let build s = minimize (X.build s)
end



module X: AutomataIface = struct
  type v = env
  let build syst = build syst
  let sizealpha syst = 1 lsl (syst.nbparam + syst.nbvar)
  let v_to_int = env_to_int
end

module Minimized = Minimizer(X)

let print_tab out t =
  let rec repeat s = function
    | 0 -> ()
    | k -> output_string out s; repeat s (k/2)
  in
  let pb i n =
    let k = (i-1) land (lnot i) land (n-1) in
    if (i > 0) then begin
      repeat "}" k; output_string out ", "
    end;
    if (i < n) then
      repeat "{" k;
  in
  for i=0 to (Array2.dim1 t)-1 do
    Printf.fprintf out "  /* %2i */ " i;
    for j=0 to (Array2.dim2 t)-1 do
      pb j (Array2.dim2 t); Printf.fprintf out "%2i" (msign t.{i,j});
    done;
    pb (Array2.dim2 t) (Array2.dim2 t);
    output_string out "\n";
  done
  
let dump_tab out t w =
  for i=0 to (Array2.dim1 t)-1 do
    for j=0 to (Array2.dim2 t)-1 do
      for s=0 to w/8-1 do
	output_byte out (t.{i,j} lsr (8*s))
      done;
    done;
  done

let print_gv out trap t tri =
  let rec get_binary w x = match w/2 with
    | 0 -> ""
    | m -> (if ((x land m) == 0) then "0" else "1")^(get_binary m x)
  in
  let rec get_trinary w x =
    let rec f b t accu =
      if (b = w) then accu
      else
        match t mod 3 with
          | 0 -> f (3*b) (t/3) (accu^"0")
          | 1 -> f (3*b) (t/3) (accu^"1")
          | 2 -> f (3*b) (t/3) (accu^"?")
    in f 1 x ""
  in let to_str = if tri then get_trinary else get_binary
  in let rec join_list = function
    | []   -> ""
    | [s]  -> s
    | h::t -> h^"\\n"^(join_list t) in
     
     for i=0 to (Array2.dim1 t)-1 do
       let h = Hashtbl.create 16 in
       for j=0 to (Array2.dim2 t)-1 do
         let x = msign t.{i,j} in
         if (trap || (x <> -1)) then
           let l = try Hashtbl.find h x with Not_found -> [] in
           Hashtbl.remove h x;
           Hashtbl.add h x (j::l)
       done;
       
       Hashtbl.iter (fun a b ->
         Printf.fprintf out "  %i -> %i [ label = \"" i a;
         output_string out (join_list (List.map (to_str (Array2.dim2 t)) b));
         Printf.fprintf out "\" ];\n"
       ) h;
     done

module HashedSet = struct
  type t = int list
  let rec equal x y = match (x,y) with
    | [], [] -> true
    | h1::t1, h2::t2 -> h1 = h2 && equal t1 t2
    | _, _ -> false
  let hash = List.fold_left (fun x y -> (y + x)*1729+257) 1
end

let rec union l1 l2 = match l1, l2 with
  | x, [] | [], x -> x
  | h1::t1, h2::t2 when h1=h2 -> h1::(union t1 t2)
  | h1::t1, h2::t2 when h1<h2 -> h1::(union t1 l2)
  | h1::t1, h2::t2            -> h2::(union l1 t2)

let rec to_bitfield = function
  | [] -> 0
  | h::t -> (1 lsl h) lor (to_bitfield t)

let rec from_bitfield = function
  | 0  -> []
  | 1  -> [0]
  | 2  -> [1]
  | 3  -> [0;1]
  | 4  -> [2]
  | 5  -> [0;2]
  | 6  -> [1;2]
  | 7  -> [0;1;2]
  | 8  -> [3]
  | 9  -> [0;3]
  | 10 -> [1;3]
  | 11 -> [0;1;3]
  | 12 -> [2;3]
  | 13 -> [0;2;3]
  | 14 -> [1;2;3]
  | 15 -> [0;1;2;3]
  | x  -> (from_bitfield (x mod 16))@
    (List.map (fun x->x+4) (from_bitfield (x/16)))

module SetTable = Hashtbl.Make(HashedSet)


let decision syst =
  let at = Minimized.build syst in
  debug "Determinizing";
  let automata = ref [] in
  let idx      = SetTable.create 64 in
  let sets     = Hashtbl.create 64 in
  SetTable.add idx [] (-1);
  let make_set l =
    let rec cleandup = function
      | []               -> []
      | [x]              -> [x]
      | x::y::l when x=y -> cleandup (y::l)
      | x::y::l          -> x::cleandup (y::l)
    in cleandup (List.fast_sort compare l)
  in
  let rec trans cur max =
    if cur <= max then
      let t = Array.make (1 lsl (syst.nbparam)) [] in
      let tt = Array.make (1 lsl (syst.nbparam)) (-1) in
      automata := tt::!automata;
      let int_params x =
        (x lsr syst.nbvar) in
      let f x = 
        let g u v = 
          if v <> -1 then
            t.(int_params u) <- v::t.(int_params u);
        in
        if x <> -1 then
	  for j=0 to (Array2.dim2 at)-1 do
            g j (msign at.{x,j});
	  done;
      in
      List.iter f (Hashtbl.find sets cur);
      debug ".";
      Array.iteri (fun i x -> t.(i) <- make_set x) t;
      let max2 = Array.fold_left (fun i v -> if SetTable.mem idx v then i else
        let m = i+1 in
        SetTable.add idx v m;
        Hashtbl.add sets m v;
        m) max t in
      Array.iteri (fun i x -> tt.(i) <- SetTable.find idx x) t;
      trans (cur+1) max2;
  in
  Hashtbl.add sets 0 [0];
  SetTable.add idx [0] 0;
  trans 0 0;
  debug "\n";
  Array2.of_array int_type Bigarray.c_layout (Array.of_list (List.rev !automata))

external determinize_c : 
  (int, int_type_elt, Bigarray.c_layout) Bigarray.Array2.t ->
    int -> (int, int_type_elt, Bigarray.c_layout) Bigarray.Array1.t
      = "determinize_c_stub"

(* external bound_c :  *)
(*   (int, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array3.t -> int *)
(*       = "bound_c_stub" *)

let decision2 syst =
  let at = Minimized.build syst in
  let use_bitfield = false && Array2.dim1 at <= 16 in
  let sz = 1 lsl syst.nbparam in
  debug "Determinizing";
  let automata = create_map_array_anon sz int_type in
  let idx      = SetTable.create 64 in
  let sets     = Hashtbl.create 64 in
  let bf_idx   = Hashtbl.create 64 in
  let bf_sets  = Hashtbl.create 64 in
  SetTable.add idx [] (-1);
  let make_set l =
    let rec cleandup = function
      | []               -> []
      | [x]              -> [x]
      | x::y::l when x=y -> cleandup (y::l)
      | x::y::l          -> x::cleandup (y::l)
    in cleandup (List.fast_sort compare l)
  in

  let base = Array.make (Array2.dim1 at)
    (Array1.create int_type Bigarray.c_layout 0) in
  let ba = if use_bitfield then
      Array2.create int_type Bigarray.c_layout
        (Array2.dim1 at) sz
    else
      Array2.create int_type Bigarray.c_layout 0 0
  in
  let bt = ref (Array1.create int_type Bigarray.c_layout 0) in
  (* let tu = Array.make sz [] in *)
  (* let ts = Array.init sz (fun _ -> SetTable.create 64) in *)
  (* let tss = Array.init sz (fun _ -> SetTable.create 64) in *)

  let rec trans cur max =
     

    if cur <= max then
      let t = Array.make sz [] in
      grow_map_array automata;
      let tt = Array2.slice_left automata.ba ((Array2.dim1 automata.ba)-1) in
     
      let f_base x = 
        let g u v = 
          if v <> -1 then
            let uu = (u lsr syst.nbvar) in
            t.(uu) <- v::t.(uu);
        in
        if x <> -1 then
          for j=0 to (Array2.dim2 at)-1 do
            g j (msign at.{x,j})
          done;
      in
      let f x = 
        if x <> -1 then
          for i=0 to (sz-1) do
            let y = msign base.(x).{i} in
            if y <> -1 then
              t.(i) <- union t.(i) (Hashtbl.find sets y)
          done;
      in
      let max2 = ref max in
      begin
        if use_bitfield then (
          if cur < Array2.dim1 at then (
            List.iter f_base (from_bitfield (Hashtbl.find bf_sets cur));
            Array.iteri (fun i x -> ba.{cur,i} <- to_bitfield (make_set x)) t;
            bt := Array2.slice_left ba cur;
            debug "#"
          ) else  (
            flush_all ();
            bt := determinize_c ba (Hashtbl.find bf_sets cur);
            debug "."
          );

          for i=0 to sz-1 do
            if Hashtbl.mem bf_idx !bt.{i} then () else (
              incr max2;
              Hashtbl.add bf_idx !bt.{i} !max2;
              Hashtbl.add bf_sets !max2 !bt.{i};
            );
            tt.{i} <- Hashtbl.find bf_idx !bt.{i};
          done;
        ) else (

          if cur < Array2.dim1 at then (
            List.iter f_base (Hashtbl.find sets cur);
            base.(cur) <- tt;
            debug "#"
          ) else (
            List.iter f (Hashtbl.find sets cur);
            debug "."
          );
            
          if cur < Array2.dim1 at then (
            Array.iteri (fun i x -> t.(i) <- make_set x) t;
            (* Array.iter (fun l -> List.iter (Printf.printf " %3i") l; print_string ",") t; *)
            (* Array.iteri (fun i x -> tu.(i) <- tu.(i)@x) t; *)
            (* Array.iteri (fun i l -> if not (SetTable.mem ts.(i) l) then SetTable.add ts.(i) (make_set l) 1) t; *)
            (* Array.iteri (fun i l -> if not (SetTable.mem tss.(i) l) then *)
            (*     SetTable.add tss.(i) l 1; *)
            (*     SetTable.iter (fun z _ -> *)
            (*       let ns = make_set (z@l) in *)
            (*       if not (SetTable.mem tss.(i) ns) then SetTable.add tss.(i) ns 1 *)
            (*     ) (SetTable.copy tss.(i)) *)
            (* ) t; *)
            (* print_string "\n"; *)
            (* flush_all (); *)
          );
          if cur = Array2.dim1 at then (
            (* print_string "\n"; *)
            (* Array.iter (fun l -> Printf.printf "%i, " (List.length (make_set l))) tu; *)
            (* print_string "\n"; *)
            (* Array.iter (fun x -> Printf.printf "%i, " (SetTable.length x)) ts; *)
            (* print_string "\n"; *)
            (* Array.iter (fun x -> Printf.printf "%i, " (SetTable.length x)) tss; *)
            (* print_string "\n"; *)
            (* print_string "\n"; *)
            (* Array.iter (fun l -> List.iter (Printf.printf " %3i") (make_set l); print_string ",") tu; *)
            (* print_string "\n"; *)
            (* flush_all (); *)
            (* let total = SetTable.create 64 in *)
            (* Array.iter (fun x -> SetTable.iter (fun z _ -> *)
            (*   if not (SetTable.mem total z) then SetTable.add total z 1 *)
            (* ) x) tss; *)
            (* debug (Printf.sprintf "\nAt most %i states\n" (SetTable.length total)); *)

            let bbase = Array3.create Bigarray.int8_unsigned Bigarray.c_layout
              (Array2.dim1 at) sz 20 in
            Array3.fill bbase 0;
            
            for i=0 to (Array2.dim1 at) - 1 do
              for j=0 to sz - 1 do
                List.iter (fun x ->
                  bbase.{i,j,x/8} <- 1 lsl (x mod 8);
                ) (if msign base.(i).{j} <> -1 then Hashtbl.find sets base.(i).{j} else [])
              done
            done;
            (* bound_c bbase; *)
            
            ()
            (* exit 0; *)
          );
          Array.iteri (fun i x -> t.(i) <- make_set x) t;
          max2 := Array.fold_left (fun i v -> if SetTable.mem idx v then i else
              let m = i+1 in
              SetTable.add idx v m;
              Hashtbl.add sets m v;
              m) max t;
          Array.iteri (fun i x -> tt.{i} <- SetTable.find idx x) t;
        );
      end;

      trans (cur+1) !max2;
  in
  for i=0 to (Array2.dim1 at)-1 do
    if use_bitfield then (
      Hashtbl.add bf_sets i (1 lsl i);
      Hashtbl.add bf_idx (1 lsl i) i
    ) else (
      Hashtbl.add sets i [i];
      SetTable.add idx [i] i
    )
  done;
  if use_bitfield then (
    Hashtbl.add bf_sets (-1) 0;
    Hashtbl.add bf_idx 0 (-1)
  );
  trans 0 ((Array2.dim1 at)-1);
  debug "\n";
  Printf.fprintf stderr "%i states\n" (Array2.dim1 automata.ba);
  debug "\n";
  automata.ba
  
(* let partial_nd at qi npar = *)
(*   debug "Partiel determinizing [01?]"; *)
(*   let automata = ref [] in *)
(*   let idx      = SetTable.create 64 in *)
(*   let sets     = Hashtbl.create 64 in *)
(*   SetTable.add idx [] (-1); *)
(*   let make_set l = *)
(*     let rec cleandup = function *)
(*       | []               -> [] *)
(*       | [x]              -> [x] *)
(*       | x::y::l when x=y -> cleandup (y::l) *)
(*       | x::y::l          -> x::cleandup (y::l) *)
(*     in cleandup (List.fast_sort compare l) *)
(*   in *)
(*   let rec trans cur max = *)
(*     if cur <= max then *)
(*       let t = Array.make ((pow3 (qi+1)) * (1 lsl (npar -(qi+1)))) [] in *)
(*       let tt = Array.make ((pow3 (qi+1)) * (1 lsl (npar -(qi+1)))) (-1) in *)
(*       automata := tt::!automata; *)
(*       let int_d x =                     (\* Same transition *\) *)
(*         let low = x mod (pow3 qi) and *)
(*             high = x / (pow3 qi) in *)
(*         low + ((high mod 2) + 3*(high/2)) * (pow3 qi) in *)
(*       let int_nd x =                    (\* 0/1 -> ? *\) *)
(*         let low = x mod (pow3 qi) and *)
(*             high = x / (pow3 qi) in *)
(*         low + (2 + 3*(high/2)) * (pow3 qi) in *)
(*       let f x =  *)
(*         let g u v =  *)
(*           if v <> -1 then *)
(*             ( t.(int_d  u) <- v::t.(int_d  u); *)
(*               t.(int_nd u) <- v::t.(int_nd u) ) *)
(*         in *)
(*         if x <> -1 then *)
(*           Array.iteri g at.(x); *)
(*       in *)
(*       List.iter f (Hashtbl.find sets cur); *)
(*       debug "."; *)
(*       Array.iteri (fun i x -> t.(i) <- make_set x) t; *)
(*       let max2 = Array.fold_left (fun i v -> if SetTable.mem idx v then i else *)
(*         let m = i+1 in *)
(*         SetTable.add idx v m; *)
(*         Hashtbl.add sets m v; *)
(*         m) max t in *)
(*       Array.iteri (fun i x -> tt.(i) <- SetTable.find idx x) t; *)
(*       trans (cur+1) max2; *)
(*   in *)
(*   Hashtbl.add sets 0 [0]; *)
(*   SetTable.add idx [0] 0; *)
(*   trans 0 0; *)
(*   debug "\n"; *)
(*   Array.of_list (List.rev !automata) *)

(*
let decision2 syst =
  let at = minimize (decision syst) in
  (* partial_nd at 0 syst.nbparam *)
  let rec f x = function
    | i when i = syst.nbparam -> x
    | i -> f (minimize (partial_nd x i syst.nbparam)) (i+1)
  in f at 0
*)

let decision3 syst =
  let at = Minimized.build syst in
  let use_bitfield = Array2.dim1 at <= 16 in
  let sz = pow3 syst.nbparam in
  debug "Determinizing";
  let automata = create_map_array_anon sz int_type in
  let idx      = SetTable.create 64 in
  let sets     = Hashtbl.create 64 in
  let bf_idx   = Hashtbl.create 64 in
  let bf_sets  = Hashtbl.create 64 in
  SetTable.add idx [] (-1);
  let make_set l =
    let rec cleandup = function
      | []               -> []
      | [x]              -> [x]
      | x::y::l when x=y -> cleandup (y::l)
      | x::y::l          -> x::cleandup (y::l)
    in cleandup (List.fast_sort compare l)
  in

  let base = Array.make (Array2.dim1 at)
    (Array1.create int_type Bigarray.c_layout 0) in
  let ba = if use_bitfield then
      Array2.create int_type Bigarray.c_layout
        (Array2.dim1 at) sz
    else
      Array2.create int_type Bigarray.c_layout 0 0
  in
  let bt = ref (Array1.create int_type Bigarray.c_layout 0) in

  let rec trans cur max =
     

    if cur <= max then
      let t = Array.make sz [] in
      grow_map_array automata;
      let tt = Array2.slice_left automata.ba ((Array2.dim1 automata.ba)-1) in
      
      let rec next = function
        | []   -> raise Over
        | 0::t -> 1::t
        | 1::t -> 0::(next t)
      in let rec make_idx uu accu = function
        | [] -> accu
        | 0::m -> make_idx (uu/2) (3*accu+(uu mod 2)) m
        | 1::m -> make_idx (uu/2) (3*accu+2) m
      in
      
      let f_base x = 
        let g u v = 
          if v <> -1 then
            let uu = bitrev syst.nbparam (u lsr syst.nbvar) in
            let rec iter mask =
              t.(make_idx uu 0 mask) <- v::t.(make_idx uu 0 mask);
                iter (next mask)
            in 
            try
              iter (Array.to_list (Array.make syst.nbparam 0))
            with Over -> ()
        in
        if x <> -1 then
          for j=0 to (Array2.dim2 at)-1 do
            g j (msign at.{x,j})
          done;
      in
      let f x = 
        if x <> -1 then
          for i=0 to (sz-1) do
            let y = msign base.(x).{i} in
            if y <> -1 then
              t.(i) <- union t.(i) (Hashtbl.find sets y)
          done;
      in
      let max2 = ref max in
      begin
        if use_bitfield then (
          if cur < Array2.dim1 at then (
            List.iter f_base (from_bitfield (Hashtbl.find bf_sets cur));
            Array.iteri (fun i x -> ba.{cur,i} <- to_bitfield (make_set x)) t;
            bt := Array2.slice_left ba cur;
            debug "#"
          ) else  (
            flush_all ();
            bt := determinize_c ba (Hashtbl.find bf_sets cur);
            debug "."
          );

          for i=0 to sz-1 do
            if Hashtbl.mem bf_idx !bt.{i} then () else (
              incr max2;
              Hashtbl.add bf_idx !bt.{i} !max2;
              Hashtbl.add bf_sets !max2 !bt.{i};
            );
            tt.{i} <- Hashtbl.find bf_idx !bt.{i};
          done;
        ) else (

          if cur < Array2.dim1 at then (
            List.iter f_base (Hashtbl.find sets cur);
            base.(cur) <- tt;
            debug "#"
          ) else (
            List.iter f (Hashtbl.find sets cur);
            debug "."
          );
            
          Array.iteri (fun i x -> t.(i) <- make_set x) t;
          max2 := Array.fold_left (fun i v -> if SetTable.mem idx v then i else
              let m = i+1 in
              SetTable.add idx v m;
              Hashtbl.add sets m v;
              m) max t;
          Array.iteri (fun i x -> tt.{i} <- SetTable.find idx x) t;
        );
      end;

      trans (cur+1) !max2;
  in
  for i=0 to (Array2.dim1 at)-1 do
    if use_bitfield then (
      Hashtbl.add bf_sets i (1 lsl i);
      Hashtbl.add bf_idx (1 lsl i) i
    ) else (
      Hashtbl.add sets i [i];
      SetTable.add idx [i] i
    )
  done;
  if use_bitfield then (
    Hashtbl.add bf_sets (-1) 0;
    Hashtbl.add bf_idx 0 (-1)
  );
  trans 0 ((Array2.dim1 at)-1);
  debug "\n";
  (* Printf.fprintf stderr "%i states\n" (Array2.dim1 automata.ba); *)
  debug "\n";
  automata.ba

    
module Y3: AutomataIface = struct
  type v = bool array
  let build syst = decision3 syst
  let sizealpha syst = 1 lsl syst.nbparam
  let v_to_int x =
    Array.fold_left (fun x b -> 2*x + if b then 1 else 0) 0 x
end

module Decision3 = Minimizer(Y3)

module Y: AutomataIface = struct
  type v = bool array
  let build syst = decision syst
  let sizealpha syst = 1 lsl syst.nbparam
  let v_to_int x =
    Array.fold_left (fun x b -> 2*x + if b then 1 else 0) 0 x
end

module Decision = Minimizer(Y)


let main () =
  let dump, gv, trans, decid, sparse, trap, tri = ref false, ref false, ref false, ref false, ref false, ref false, ref false in
  let bits = ref 5 in
  let w = ref 16 in
  let out = ref stdout in
  let lex = ref None in
  let anon s =
    lex := Some (Lexing.from_channel (open_in s)) in
  let _ = Arg.parse [
    "-v", Arg.Set verbose, "Verbose (debug output)";
    "-b", Arg.Set dump, "Dump in binary format instead of C format";
    "-w", Arg.Set_int w, "Word size for binary dump";
    "-g", Arg.Set gv, "Output in Graphviz format";
    "-dead", Arg.Set trap, "Graphviz format: print dead state";
    "-t", Arg.Set trans, "Output transition FSM";
    "-d", Arg.Set decid, "Output decision FSM";
    "-s", Arg.Set sparse, "Output sparse decision FSM (binary format)";
    "-sb", Arg.Set_int bits, "Sparse FSM parameter width (3<=sb<=5)";
    "-3", Arg.Set tri, "Output tri-valued decision FSM (0/1/?)";
    "-o", Arg.String (fun s -> out := open_out s), "Output file";
    "-i", Arg.Unit (fun x -> lex := Some (Lexing.from_channel stdin)), "Read from standard input";
    "-e", Arg.String (fun s -> lex := Some (Lexing.from_string s)), "Take expression from command line";
  ] anon "This tool build the FSM corresponding to an S-Function"
  (* in let pre = if !dump then "" else "// " *)
  in let print t s =
    if !dump then
        dump_tab !out t !w
    else if !gv then
      (
        output_string !out ("digraph finite_state_machine {\n"^
	  "rankdir = LR;\nnode [shape = circle];\n");
        print_gv !out !trap t !tri;
        output_string !out "}\n"
      )
    else
      (
        output_string !out ("int "^s^"[] = {\n");
        print_tab !out t;
        output_string !out "};\n"
      )
  in match !lex with
    | None -> print_string "No input given!\n"
    | Some lexbuf ->
        let a = try Parser.system Lexer.token lexbuf
            with exn ->
              begin
                let curr = lexbuf.Lexing.lex_curr_p in
                (* let line = curr.Lexing.pos_lnum in *)
                (* let cnum = curr.Lexing.pos_cnum - curr.Lexing.pos_bol in *)
                let tok = Lexing.lexeme lexbuf in
                (* let tail = Lexer.ruleTail "" lexbuf in *)
                failwith (Printf.sprintf "Parse failure: char %i [token %s]\n" curr.Lexing.pos_cnum tok)
              end
        in
        debug (Printf.sprintf "Parsed expression: %i params, %i vars, %i sums\n" a.nbparam a.nbvar a.nbsum);
        flush_all ();
        if !trans then (
          print (Minimized.build a) "transitions"
        );
        if !decid then (
          print (Decision.build a) "decision"
        );
        if !sparse then (
          build_sparse a (Unix.descr_of_out_channel !out) !bits
        );
	if !tri then (
          print (Decision3.build a) "decision3"
	);
        if not (!decid or !trans or !tri or !sparse) then
          print_string "No output asked.  Please use -d or -t (or both).\n"
        
let _ = main ()
