/* 
 *  ARXtools - A toolkit for ARX analysis
 *
 *  Copyright: Gaëtan Leurent 2011-2012 <gaetan.leurent@normalesup.org>
 *
 *  To the extent possible under law, the author(s) have dedicated all
 *  copyright and related and neighboring rights to this software to the
 *  public domain worldwide. This software is distributed without any
 *  warranty.
 *
 *  You should have received a copy of the CC0 Public Domain Dedication
 *  along with this software. If not, see
 *  <http://creativecommons.org/publicdomain/zero/1.0/>.
 */


/* 
 * This tool is used to manipulate our new set of constraints
 * for differential analysis of ARX ciphers and hash functions.
 * 
 * You should first use build_fsm to construct the automata
 * for each operation used if the cipher.  Then use this tool
 * to compute the probabilities for each operation, and to to
 * propagate constraints.
 */

/*
 * WARNING:
 * Make sure that the automaton use the correct integer types!
 * The type used in this program is defined as state_t, modify it if needed.
 */


#define _XOPEN_SOURCE 700

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <assert.h>
#include <setjmp.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef _WIN32
#include <sys/mman.h>
#else
#include "mmap_windows.c"
#endif

#define tabsize(t) (sizeof(t)/sizeof((t)[0]))

int pow3(int n) {
  if (n==0)
    return 1;
  else
    return 3*pow3(n-1);
}

typedef int16_t state_t;

const state_t diff_count[][2][2][2][2][2][2] = {
  /*  0 */ {{{{{{ 0, -1}, {-1,  1}}, {{ 0, -1}, {-1, -1}}}, {{{ 0, -1}, {-1, -1}}, {{ 0, -1}, { 1, -1}}}}, {{{{ 0,  0}, { 1,  1}}, {{-1, -1}, {-1,  1}}}, {{{-1, -1}, {-1,  1}}, {{-1,  0}, {-1,  1}}}}}, {{{{{-1,  0}, { 1, -1}}, {{-1,  0}, {-1, -1}}}, {{{-1,  0}, {-1, -1}}, {{ 0,  0}, {-1, -1}}}}, {{{{-1, -1}, {-1, -1}}, {{-1, -1}, { 1, -1}}}, {{{-1, -1}, { 1, -1}}, {{-1, -1}, { 1,  1}}}}}}, 
  /*  1 */ {{{{{{ 0, -1}, {-1,  1}}, {{ 0, -1}, {-1, -1}}}, {{{-1, -1}, {-1,  1}}, {{ 0, -1}, { 1, -1}}}}, {{{{ 0,  0}, { 1,  1}}, {{-1, -1}, {-1,  1}}}, {{{ 0, -1}, {-1, -1}}, {{-1,  0}, {-1,  1}}}}}, {{{{{-1,  0}, { 1, -1}}, {{-1,  0}, {-1, -1}}}, {{{-1, -1}, { 1, -1}}, {{ 0,  0}, {-1, -1}}}}, {{{{-1, -1}, {-1, -1}}, {{-1, -1}, { 1, -1}}}, {{{-1,  0}, {-1, -1}}, {{-1, -1}, { 1,  1}}}}}}, 
};


#ifdef DEBUG
#define print_tab2(t,n) do {                            \
    printf ("Table %s:\n", #t);                         \
    for (unsigned int i__=0; i__<n; i__++)              \
      printf ("  %08x", t[i__]);                        \
    printf ("\n");                                      \
  } while(0)

#define print_var(v)                            \
  printf("Var %s: %08x\n", #v, v)

#define log_var(v,log) do {                     \
    printf("Var %s: %08x\n", #v, v);            \
    *(log++) = v;                               \
  } while(0)
#else
#define print_tab2(t,n)
#define print_var(v)
#define log_var(v,log)
#endif

#define print_tab(t) print_tab2(t, tabsize(t))


#define rotl32(x,n)   (((x) << (n)) | ((x) >> (32 - (n))))
#define rotr32(x,n)   (((x) >> (n)) | ((x) << (32 - (n))))

#define rotl64(x,n)   (((x) << (n)) | ((x) >> (64 - (n))))
#define rotr64(x,n)   (((x) >> (n)) | ((x) << (64 - (n))))


unsigned long long cnt = 0;
int best_s2 = -1;
unsigned int s2_stat[33] = {0};

int best_s1 = -1;
unsigned int s1_stat[33] = {0};

double cost;

#define MAX_PARAMS 12  // Increase if needed

typedef struct {
  struct reg *r;
  unsigned int rot;
} parameter;

typedef uint64_t word;
#define MAX_WORD_SIZE ((int)(8*sizeof(word)))
int WORD_SIZE = 32;

typedef struct {
  state_t *data;
  int nb_states;
  int nb_params;
  int nb_vars;
} automaton_t;

typedef struct {
  parameter p[MAX_PARAMS];
  automaton_t transition;
  automaton_t decision;
  automaton_t decision3;
} equation;

typedef struct reg {
  char bits[MAX_WORD_SIZE];
} reg;

typedef struct {
  unsigned int bad;
  unsigned int nb;
  state_t prev;
  char c;
} path_data;

// typedef used for counting solutions
// should allows relatively large numbers with good precision
typedef long double bignum;
// Recent GCC also support __float128
// typedef __float128 bignum;

typedef struct {
  bignum count;
} path_data_proba;


// Describe the set of constraints

#define CSTR_BITS 4
#define CSTR_MAX (1<<CSTR_BITS)

enum constraints {
  CONSTR_EQ     = 0,
  CONSTR_NEQ    = 8,
  CONSTR_UNKOWN = 4,
  CONSTR_SAME   = 2,
  CONSTR_XSAME  = 10,
  CONSTR_DIFF   = 6,
  CONSTR_XDIFF  = 14,
  CONSTR_ZERO   = 1,
  CONSTR_UP     = 9,
  CONSTR_ONE    = 5,
  CONSTR_DOWN   = 13,
  CONSTR_X0     = 11,
  CONSTR_X1     = 15,
  CONSTR_Y0     = 3,
  CONSTR_Y1     = 7,
};

char constr_to_char[CSTR_MAX] = {
  [CONSTR_EQ    ] = '-',
  [CONSTR_NEQ   ] = 'x',
  [CONSTR_UNKOWN] = '?',
  [CONSTR_SAME  ] = '=',
  [CONSTR_XSAME ] = '<',
  [CONSTR_DIFF  ] = '!',
  [CONSTR_XDIFF ] = '>',
  [CONSTR_ZERO  ] = '0',
  [CONSTR_UP    ] = 'u',
  [CONSTR_ONE   ] = '1',
  [CONSTR_DOWN  ] = 'n',
  [CONSTR_X0    ] = '3',
  [CONSTR_X1    ] = 'C',
  [CONSTR_Y0    ] = '5',
  [CONSTR_Y1    ] = 'A',
};

enum constraints constr_rotation_safe[CSTR_MAX] = {
  [CONSTR_EQ    ] = CONSTR_EQ,
  [CONSTR_NEQ   ] = CONSTR_NEQ,
  [CONSTR_UNKOWN] = CONSTR_UNKOWN,
  [CONSTR_SAME  ] = CONSTR_EQ,
  [CONSTR_XSAME ] = CONSTR_NEQ,
  [CONSTR_DIFF  ] = CONSTR_EQ,
  [CONSTR_XDIFF ] = CONSTR_NEQ,
  [CONSTR_ZERO  ] = CONSTR_ZERO,
  [CONSTR_UP    ] = CONSTR_UP,
  [CONSTR_ONE   ] = CONSTR_ONE,
  [CONSTR_DOWN  ] = CONSTR_DOWN,
  [CONSTR_X0    ] = CONSTR_X0,
  [CONSTR_X1    ] = CONSTR_X1,
  [CONSTR_Y0    ] = CONSTR_Y0,
  [CONSTR_Y1    ] = CONSTR_Y1,
};

#define CONSTR_HAS_DIFF(c) constr_has_diff[c]

int constr_has_diff[CSTR_MAX] = {
  [CONSTR_EQ    ] = 0,
  [CONSTR_NEQ   ] = 1,
  [CONSTR_UNKOWN] = 0,  // Unknown
  [CONSTR_SAME  ] = 0,
  [CONSTR_XSAME ] = 1,
  [CONSTR_DIFF  ] = 0,
  [CONSTR_XDIFF ] = 1,
  [CONSTR_ZERO  ] = 0,
  [CONSTR_UP    ] = 1,
  [CONSTR_ONE   ] = 0,
  [CONSTR_DOWN  ] = 1,
  [CONSTR_X0    ] = 0,  // Unknown
  [CONSTR_X1    ] = 0,  // Unknown
  [CONSTR_Y0    ] = 0,  // Unknown
  [CONSTR_Y1    ] = 0,  // Unknown
};

int constr_to_xdiff[CSTR_MAX] = {
  [CONSTR_EQ    ] = 0,
  [CONSTR_NEQ   ] = 1,
  [CONSTR_UNKOWN] = 2,  // Unknown
  [CONSTR_SAME  ] = 0,
  [CONSTR_XSAME ] = 1,
  [CONSTR_DIFF  ] = 0,
  [CONSTR_XDIFF ] = 1,
  [CONSTR_ZERO  ] = 0,
  [CONSTR_UP    ] = 1,
  [CONSTR_ONE   ] = 0,
  [CONSTR_DOWN  ] = 1,
  [CONSTR_X0    ] = 2,  // Unknown
  [CONSTR_X1    ] = 2,  // Unknown
  [CONSTR_Y0    ] = 2,  // Unknown
  [CONSTR_Y1    ] = 2,  // Unknown
};

enum constraints constr_to_active[CSTR_MAX] = {
  [CONSTR_EQ    ] = CONSTR_EQ,
  [CONSTR_NEQ   ] = CONSTR_NEQ,
  [CONSTR_UNKOWN] = CONSTR_NEQ,
  [CONSTR_SAME  ] = CONSTR_SAME,
  [CONSTR_XSAME ] = CONSTR_XSAME,
  [CONSTR_DIFF  ] = CONSTR_DIFF,
  [CONSTR_XDIFF ] = CONSTR_XDIFF,
  [CONSTR_ZERO  ] = CONSTR_ZERO,
  [CONSTR_UP    ] = CONSTR_UP,
  [CONSTR_ONE   ] = CONSTR_ONE,
  [CONSTR_DOWN  ] = CONSTR_DOWN,
  [CONSTR_X0    ] = CONSTR_UP,
  [CONSTR_X1    ] = CONSTR_DOWN,
  [CONSTR_Y0    ] = CONSTR_DOWN,
  [CONSTR_Y1    ] = CONSTR_UP,
};

enum constraints constr_to_inactive[CSTR_MAX] = {
  [CONSTR_EQ    ] = CONSTR_EQ,
  [CONSTR_NEQ   ] = CONSTR_NEQ,
  [CONSTR_UNKOWN] = CONSTR_EQ,
  [CONSTR_SAME  ] = CONSTR_SAME,
  [CONSTR_XSAME ] = CONSTR_XSAME,
  [CONSTR_DIFF  ] = CONSTR_DIFF,
  [CONSTR_XDIFF ] = CONSTR_XDIFF,
  [CONSTR_ZERO  ] = CONSTR_ZERO,
  [CONSTR_UP    ] = CONSTR_UP,
  [CONSTR_ONE   ] = CONSTR_ONE,
  [CONSTR_DOWN  ] = CONSTR_DOWN,
  [CONSTR_X0    ] = CONSTR_ZERO,
  [CONSTR_X1    ] = CONSTR_ONE,
  [CONSTR_Y0    ] = CONSTR_ZERO,
  [CONSTR_Y1    ] = CONSTR_ONE,
};

void read_reg_from_str (reg *r, const char* s) {
  static char *to_const = NULL;
  if (!to_const) {
    to_const = malloc(256);
    memset (to_const, -1, 256);
    for (int i=0; i<CSTR_MAX; i++)
      if (constr_to_char[i])
        to_const[constr_to_char[i]] = i;
  }

  int i=0;
  for (; *s && i<WORD_SIZE; s++) {
    int c = to_const[*s];
    if (c == -1) {
      printf ("Warning: Unknown char: `%c'\n", *s);
      c = CONSTR_UNKOWN;
    }
    r->bits[WORD_SIZE-1-i++] = c;
  }

  if (*s)
    printf ("Warning: string to long\n");

  if (i < WORD_SIZE) {
    printf ("Warning: string to short\n");
    for (; i<WORD_SIZE; i++)
      r->bits[i] = CONSTR_UNKOWN;
  }
}

void print_reg(reg *r) {
  for (int i=WORD_SIZE-1; i>=0; i--) {
    printf ("%c", constr_to_char[r->bits[i]]);
  }
}



/*
 * Returns the first failed bit.
 * Returns WORD_SIZE if system is actually compatible.
 *
 * Uses the decision automaton for better performance.
 */

int eq_compatible_upto(equation *eq) {

  parameter *in = eq->p;

  /* for (int i=0; i<npar; i++) { */
  /*   print_reg(in[i].r); printf (" "); */
  /* } */
  /* printf ("\n"); */

  state_t state = 0;

  if (eq->decision.data) {

    automaton_t *at = &eq->decision;
    int npar = at->nb_params;
    state_t (*aa)[1<<(CSTR_BITS*npar)] = (void*) at->data;
    
    int i;
    for (i=0; i < WORD_SIZE; i++) {
      unsigned int par = 0;
      for (int k=0; k<npar; k++) {
        par <<= CSTR_BITS;
        int x = in[k].r->bits[(i+in[k].rot) % WORD_SIZE];
        if (i == 0)
          x = constr_rotation_safe[x];
        par  += x;
      }
      /* printf ("%4hx,%03x -> %4hx [%x]\n", state, par, aa[state][par], */
      /*         (char*)&aa[state][par] - (char*)aa); */
      state = aa[state][par];
      if ((unsigned) state == (unsigned) -1) {
        return i;
      }
    }
    
    return i;

  } else if (eq->decision3.data) {

    automaton_t *at = &eq->decision3;
    int npar = at->nb_params;
    state_t (*aa)[pow3(npar)] = (void*) at->data;
    
    int i;
    for (i=0; i < WORD_SIZE; i++) {
      unsigned int par = 0;
      for (int k=0; k<npar; k++) {
        par *= 3;
        int x = in[k].r->bits[(i+in[k].rot) % WORD_SIZE];
        if (i == 0)
          x = constr_rotation_safe[x];
        par  += constr_to_xdiff[x];
      }
      /* printf ("%3i,%05x (%s) -> %3i [%x]\n", state, par, decode(par), */
      /*         aa[state][par], */
      /*         (char*)&aa[state][par] - (char*)aa); */
      state = aa[state][par];
      if ((unsigned) state == (unsigned) -1) {
        return i;
      }
    }
    
    return i;
    
    
  } else {
    fprintf (stderr, "Error: decision automaton not defined.\n");
    return -1;
  }
}

bignum eq_count(equation *eq) {

  automaton_t *at = &eq->transition;
  if (at->data == NULL) {
    fprintf (stderr, "Error: transition automaton not defined.\n");
    return -1;
  }

  int npar = at->nb_params;
  int nvar = at->nb_vars;
  int nst  = at->nb_states;
  parameter *in = eq->p;

  state_t (*aa)[1<<(CSTR_BITS*npar)][1<<nvar] = (void*) at->data;
  path_data_proba p[WORD_SIZE+1][nst];
  memset (p, 0, sizeof(p));
  p[0][0].count = 1;
  for (int i=0; i<WORD_SIZE; i++) {
    unsigned int par = 0;
    for (int k=0; k<npar; k++) {
      par <<= CSTR_BITS;
      int x = in[k].r->bits[(i+in[k].rot) % WORD_SIZE];
      if (i == 0)
        x = constr_rotation_safe[x];
      par  += x;
    }
    for (int j=0; j<nst; j++) {
      if (p[i][j].count) {
	for (int zz = 0; zz < (1<<nvar); zz++) {
	  state_t z = aa[j][par][zz];
	  if (z != -1) {
	    path_data_proba *pp= &p[i+1][z];
            pp->count += p[i][j].count;
	  }
	}
      }
    }
  }

  bignum ret = 0;
  for (int i=0; i<nst; i++)
    ret += p[WORD_SIZE][i].count;

  return ret;
}

int eq_compatible(equation *eq) {
  if (eq->decision.data || eq->decision3.data)
    return eq_compatible_upto(eq) == WORD_SIZE;
  else
    return eq_count(eq)? 1: 0;    
}

int propagate_eq_reg (equation *eq, reg *r, int ri) {
  int new = 0;
  assert(eq_compatible(eq));

#define pcstr(x) printf ("Sign constraint: %i.%i:%c\n", ri, i, constr_to_char[x]);

  if (eq->decision.data || (eq->transition.data && !eq->decision3.data)) {

    for (int i=0; i<WORD_SIZE; i++) {

#define test(a,b) do {                                                  \
        /* printf ("test '%c' -> '%c'/'%c'\n",                            */ \
        /* constr_to_char[r->bits[i]], constr_to_char[a],constr_to_char[b]);*/ \
        enum constraints c = r->bits[i];                                \
        r->bits[i] = a;                                                 \
        if (!eq_compatible(eq)) {                                       \
          new++;                                                        \
          pcstr (b);                                                    \
          r->bits[i] = b;                                               \
          assert(eq_compatible(eq));                                    \
          break;                                                        \
        }                                                               \
        r->bits[i] = b;                                                 \
        if (!eq_compatible(eq)) {                                       \
          new++;                                                        \
          pcstr (a);                                                    \
          r->bits[i] = a;                                               \
          assert(eq_compatible(eq));                                    \
          break;                                                        \
        }                                                               \
        r->bits[i] = c;                                                 \
      } while (0)
    

      switch (r->bits[i]) {
      case CONSTR_UNKOWN:
        test (CONSTR_EQ, CONSTR_NEQ);
        test (CONSTR_X0, CONSTR_X1);
        test (CONSTR_Y0, CONSTR_Y1);
        break;
      case CONSTR_EQ:
        test (CONSTR_SAME, CONSTR_DIFF);
      case CONSTR_SAME:
      case CONSTR_DIFF:
        test (CONSTR_ZERO, CONSTR_ONE);
        break;
      case CONSTR_NEQ:
        test (CONSTR_XSAME, CONSTR_XDIFF);
      case CONSTR_XSAME:
      case CONSTR_XDIFF:
        test (CONSTR_UP, CONSTR_DOWN);
        break;
      case CONSTR_X0:
        test(CONSTR_ZERO, CONSTR_UP);
        break;
      case CONSTR_X1:
        test(CONSTR_ONE, CONSTR_DOWN);
        break;
      case CONSTR_Y0:
        test(CONSTR_ZERO, CONSTR_DOWN);
        break;
      case CONSTR_Y1:
        test(CONSTR_ONE, CONSTR_UP);
        break;
      }
    }

#undef test

    return new;

  } else if (eq->decision3.data) {

    for (int i=0; i<WORD_SIZE; i++) {

      if (constr_to_xdiff[r->bits[i]] == 2) {
        enum constraints c = r->bits[i];
        r->bits[i] = constr_to_active[c];
        if (!eq_compatible(eq)) {
          new++;
          pcstr (constr_to_inactive[c]);
          r->bits[i] = constr_to_inactive[c];
          assert(eq_compatible(eq));
          continue;
        }
        r->bits[i] = constr_to_inactive[c];
        if (!eq_compatible(eq)) {
          new++;
          pcstr (constr_to_active[c]);
          r->bits[i] = constr_to_active[c];
          assert(eq_compatible(eq));
          continue;
        }
        r->bits[i] = c;
      }
    }

    return new;

  } else {
    fprintf (stderr, "Error: decision automaton not defined.\n");
    return -1;
  }

#undef pcstr

}


int propagate_eq (equation *eq) {
  if (!eq_compatible(eq))
    return -1;

  int new = 0;
  for (int i=0; i < eq->decision.nb_params; i++) {
    new += propagate_eq_reg (eq, eq->p[i].r, i);
  }

  if (new > 0)
    new += propagate_eq(eq);

  return new;
}

int mmap_at (char *file, automaton_t *at) {
    int fd = open(file, O_RDONLY);

    if (fd < 0) {
      char buffer[32];
      snprintf (buffer, sizeof(buffer), "open %s", file);
      perror (buffer);
      return -1;
    }
    
    struct stat st;
    fstat(fd, &st);
    size_t size = st.st_size;
    at->data = mmap(NULL, size,
                PROT_READ, MAP_SHARED, fd, 0);
    if (at->data == MAP_FAILED) {
      perror ("mmap:");
      return -1;
    }
    at->nb_states = size;
    return 0;
}


int main (int argc, char* argv[]) {
  int propagate = 0;
  int cost = 0;

  int opt;

  int nvar = -1;
  int npar = -1;
  equation e;
  memset(&e, 0, sizeof(e));

  while ((opt = getopt(argc, argv, "w:vpcn:m:d:t:3:")) != -1) {

    switch (opt) {
    case 'v':
      propagate = 0;
      cost = 0;
      break;
    case 'p':
      propagate = 1;
      break;
    case 'c':
      cost = 1;
      break;

    case 'n':
      nvar = atoi(optarg);
      break;
    case 'm':
      npar = atoi(optarg);
      break;

    case 't':
      if (mmap_at(optarg, &e.transition))
        return -1;
      break;
    case 'd':
      if (mmap_at(optarg, &e.decision))
        return -1;
      break;
    case '3':
      if (mmap_at(optarg, &e.decision3))
        return -1;
      break;

    case 'w':
      WORD_SIZE = atoi(optarg);
      break;

    default:
    case '?':
    usage:
      printf ("Usage: %s [-w n] [-v|-p|-c] [-d file] [-t file] [-n n] [-m n]\n"
              "       -- cst1 [cst2 ...]\n"
              "\n"
              "    -w <n>: set word size\n"
              "\n"
              "    -v: only verify constraints (default) [requires -d]\n"
              "    -p: propagate constraints             [requires -d]\n"
              "    -c: count solutions                   [requires -t]\n"
              "\n"
              "    -d <file>: decision automaton\n"
              "    -3 <file>: tri-valued decision automaton\n"
              "    -t <file>: transition automaton       [requires -n]\n"
              "\n"
              "    -n <n>: number of variables\n"
              "    -m <n>: number of parameters\n", argv[0]);
    return -1;
    }
  }

  // Sanity checks 

  if (e.decision.data == NULL && e.transition.data == NULL && e.decision3.data == NULL && !cost) {
    fprintf(stderr, "Automaton needed (use -t/-d/-d3 option)\n");
    return -1;
  }
  if (e.decision.data != NULL && e.decision3.data != NULL) {
    fprintf(stderr, "Bad arguments (-d and -3 contradict)\n");
    return -1;
  }
  if (e.transition.data == NULL && cost) {
    fprintf(stderr, "Transition automaton needed (use -t option)\n");
    return -1;
  }

  if (cost && nvar == -1) {
    fprintf(stderr, "Number of variable needed (use -n option)\n");
    return -1;
  }

  if (argc == optind)
    goto usage;

  if (npar != -1) {
    if (npar != argc - optind)
      goto usage;
  } else {
    npar = argc - optind;
  }

  int dline = (( 1<<(npar*CSTR_BITS) ) * sizeof(state_t));
  assert((e.decision.nb_states % dline) == 0);
  e.decision.nb_states /= dline;
  e.decision.nb_params  = npar;
  e.decision.nb_vars    = nvar;

  int dline3 = (( pow3(npar) ) * sizeof(state_t));
  assert((e.decision3.nb_states % dline3) == 0);
  e.decision3.nb_states /= dline3;
  e.decision3.nb_params  = npar;
  e.decision3.nb_vars    = nvar;
  
  int tline = (( 1<<(npar*CSTR_BITS+nvar) ) * sizeof(state_t));
  assert((e.transition.nb_states % tline) == 0);
  e.transition.nb_states /= tline;
  e.transition.nb_params  = npar;
  e.transition.nb_vars    = nvar;

  reg regs[MAX_PARAMS];

  for (int i=0; i<argc-optind; i++) {
    e.p[i].r = &regs[i];
    e.p[i].rot = 0;
    read_reg_from_str(&regs[i], argv[optind+i]);
  }

  /* print_reg(&regs[0]); printf (" "); */
  /* print_reg(&regs[1]); printf (" "); */
  /* print_reg(&regs[2]); printf ("\n"); */

  if (!eq_compatible(&e)) {
    if (e.decision.data || e.decision3.data)
      printf ("System fails at %2i\n", eq_compatible_upto(&e));
    else
      printf ("System fails\n");
  } else {
    printf ("System is compatible!\n");

    if (cost) {
      bignum p0 = eq_count(&e);
      printf ("#Sol: %llu (2^%.1f)\n", (unsigned long long)p0, log((double) p0)/log(2));

      if (npar > 1) {

        // Count the number of possible values for each parameter
        bignum counts[npar];
        bignum prod = 1;
        for (int i=0; i<npar; i++) {
          equation ee = {
            { e.p[i] },
            { (state_t*)diff_count, 2, 1, 2 },
            { NULL, 0, 0, 0 },
            { NULL, 0, 0, 0 }
          };
          counts[i] = eq_count(&ee);
          prod *= counts[i];
        }
        

        // Now compute probabilities with a random choice a a given parameter
        
        for (int i=0; i<npar; i++) {
          printf ("Cost[%i]: %.1f\n", i,
                  -log((double)(p0*counts[i]/prod))/log(2));
        }
      }
    }

    if (propagate) {
      printf ("Propagate:\n");
      int x = propagate_eq(&e);
      printf ("%i new constraints\nNew parameters:", x);
      for (int i=0; i<npar; i++) {
        printf (" "); print_reg(&regs[i]);
      }
      printf ("\n");
    }
  }

  return 0;
}
