/* 
 *  ARXtools - A toolkit for ARX analysis
 *
 *  Copyright: Gaëtan Leurent 2011-2012 <gaetan.leurent@normalesup.org>
 *
 *  To the extent possible under law, the author(s) have dedicated all
 *  copyright and related and neighboring rights to this software to the
 *  public domain worldwide. This software is distributed without any
 *  warranty.
 *
 *  You should have received a copy of the CC0 Public Domain Dedication
 *  along with this software. If not, see
 *  <http://creativecommons.org/publicdomain/zero/1.0/>.
 */


/* 
 * This tool is used to manipulate our new set of constraints
 * for differential analysis of ARX ciphers and hash functions.
 * 
 * You should first use build_fsm to construct the automata
 * for each operation used if the cipher.  Then use this tool
 * to compute the probabilities for each operation, and to to
 * propagate constraints.
 */

/*
 * WARNING:
 * Make sure that the automaton use the correct integer types!
 * The type used in this program is defined as state_t, modify it if needed.
 */


#define _XOPEN_SOURCE 700

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <assert.h>
#include <setjmp.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef _WIN32
#include <sys/mman.h>
#else
#include "mmap_windows.c"
#endif

#define tabsize(t) (sizeof(t)/sizeof((t)[0]))
#define MIN(a,b) ((a)<(b)?(a):(b))

typedef int16_t state_t;

#ifdef DEBUG
#define print_tab2(t,n) do {                            \
    printf ("Table %s:\n", #t);                         \
    for (unsigned int i__=0; i__<n; i__++)              \
      printf ("  %08x", t[i__]);                        \
    printf ("\n");                                      \
  } while(0)

#define print_var(v)                            \
  printf("Var %s: %08x\n", #v, v)

#define log_var(v,log) do {                     \
    printf("Var %s: %08x\n", #v, v);            \
    *(log++) = v;                               \
  } while(0)
#else
#define print_tab2(t,n)
#define print_var(v)
#define log_var(v,log)
#endif

#define print_tab(t) print_tab2(t, tabsize(t))


#define rotl32(x,n)   (((x) << (n)) | ((x) >> (32 - (n))))
#define rotr32(x,n)   (((x) >> (n)) | ((x) << (32 - (n))))

#define rotl64(x,n)   (((x) << (n)) | ((x) >> (64 - (n))))
#define rotr64(x,n)   (((x) >> (n)) | ((x) << (64 - (n))))


unsigned long long cnt = 0;
int best_s2 = -1;
unsigned int s2_stat[33] = {0};

int best_s1 = -1;
unsigned int s1_stat[33] = {0};

double cost;

#define MAX_PARAMS 12  // Increase if needed

typedef struct {
  struct reg *r;
  unsigned int rot;
} parameter;

typedef uint64_t word;
#define MAX_WORD_SIZE ((int)(8*sizeof(word)))
int WORD_SIZE = 32;

#define CSTR_BITS 32
typedef uint32_t state_bf;

typedef struct {
  uint32_t nb_states;
  uint32_t nb_params;
  uint32_t sparse_size;
  state_bf data[];
} sparse_automaton_t;

typedef struct {
  parameter p[MAX_PARAMS];
  sparse_automaton_t *transition_sparse;
} equation;

typedef struct reg {
  state_bf bits[MAX_WORD_SIZE];
} reg;

typedef struct {
  unsigned int bad;
  unsigned int nb;
  state_t prev;
  char c;
} path_data;

// typedef used for counting solutions
// should allows relatively large numbers with good precision
// typedef long double bignum;
// Recent GCC also support __float128
typedef __float128 bignum;

typedef struct {
  bignum count;
} path_data_proba;

sparse_automaton_t xx_count = {
  8, 1, 4,
  {
  /*  0*/ /*{*/ /*{*/1U<<31,   0/*}*/, /*{*/1U<<23,   3/*}*/, /*{*/1U<<15,   1/*}*/, /*{*/1U<< 7,   4/*}*/, /*},*/
  /*  1*/ /*{*/ /*{*/1U<<27,   2/*}*/, /*{*/1U<<19,   6/*}*/, /*{*/1U<<11,   5/*}*/, /*{*/1U<< 3,   7/*}*/, /*},*/
  /*  2*/ /*{*/ /*{*/1U<<30,   0/*}*/, /*{*/1U<<22,   3/*}*/, /*{*/1U<<14,   1/*}*/, /*{*/1U<< 6,   4/*}*/, /*},*/
  /*  3*/ /*{*/ /*{*/1U<<29,   0/*}*/, /*{*/1U<<21,   3/*}*/, /*{*/1U<<13,   1/*}*/, /*{*/1U<< 5,   4/*}*/, /*},*/
  /*  4*/ /*{*/ /*{*/1U<<25,   2/*}*/, /*{*/1U<<17,   6/*}*/, /*{*/1U<< 9,   5/*}*/, /*{*/1U<< 1,   7/*}*/, /*},*/
  /*  5*/ /*{*/ /*{*/1U<<26,   2/*}*/, /*{*/1U<<18,   6/*}*/, /*{*/1U<<10,   5/*}*/, /*{*/1U<< 2,   7/*}*/, /*},*/
  /*  6*/ /*{*/ /*{*/1U<<28,   0/*}*/, /*{*/1U<<20,   3/*}*/, /*{*/1U<<12,   1/*}*/, /*{*/1U<< 4,   4/*}*/, /*},*/
  /*  7*/ /*{*/ /*{*/1U<<24,   2/*}*/, /*{*/1U<<16,   6/*}*/, /*{*/1U<< 8,   5/*}*/, /*{*/1U<< 0,   7/*}*/, /*},*/
  }
};

// Describe the set of constraints

enum constraints {
  CONSTR_IMPOSSIBLE = 0,
  CONSTR_ONE    = 0x000000ff,
  CONSTR_DOWN   = 0x0000ff00,
  CONSTR_XC     = 0x0000ffff,
  CONSTR_UP     = 0x00ff0000,
  CONSTR_XA     = 0x00ff00ff,
  CONSTR_NEQ    = 0x00ffff00,
  CONSTR_XE     = 0x00ffffff,
  CONSTR_ZERO   = 0xff000000,
  CONSTR_EQ     = 0xff0000ff,
  CONSTR_X5     = 0xff00ff00,
  CONSTR_XD     = 0xff00ffff,
  CONSTR_X3     = 0xffff0000,
  CONSTR_XB     = 0xffff00ff,
  CONSTR_X7     = 0xffffff00,
  CONSTR_UNKOWN = 0xffffffff,

  CONSTR_SAME   = 0xf000000f,
  CONSTR_DIFF   = 0x0f0000f0,
  CONSTR_XSAME  = 0x00f00f00,
  CONSTR_XDIFF  = 0x000ff000,

  CONSTR_X      = 0xff3c3cff,
  CONSTR_CARRY  = 0xdb2424db,
  CONSTR_Y      = 0x24181824,
  CONSTR_UCARRY = 0xcf3030cf,
  CONSTR_NCARRY = 0xf30c0cf3,

  CONSTR_NOTU   = ~CONSTR_UCARRY, // 0x30cfcf30
  CONSTR_NOTN   = ~CONSTR_NCARRY, // 0x0cf3f30c

  CONSTR_I      = 0x3cffff3c,
  CONSTR_ICARRY = 0x24dbdb24,
  CONSTR_II     = CONSTR_I & ~CONSTR_ICARRY,

  CONSTR_J      = 0xffc3c3ff,

  CONSTR_P_ZERO = 0xc0c0c0c0,  // VS SAME !!!!!
  CONSTR_P_ONE  = 0x03030303,
  CONSTR_P_DOWN = 0x0c0c0c0c,
  CONSTR_P_UP   = 0x30303030,

  CONSTR_P_EQ   = 0xc3c3c3c3,
  CONSTR_P_NEQ  = 0x3c3c3c3c,
  
  CONSTR_PP_0   = 0xaaaaaaaa,
  CONSTR_PP_1   = 0x55555555,
};

char* constr_to_char(uint32_t x) {
  switch (x) {
  case CONSTR_EQ:      return  "-";
  case CONSTR_NEQ:     return  "x";
  case CONSTR_UNKOWN:  return  "?";
  case CONSTR_SAME:    return  "=";
  case CONSTR_XSAME:   return  "<";
  case CONSTR_DIFF:    return  "!";
  case CONSTR_XDIFF:   return  ">";
  case CONSTR_ZERO:    return  "0";
  case CONSTR_UP:      return  "u";
  case CONSTR_ONE:     return  "1";
  case CONSTR_DOWN:    return  "n";
  case CONSTR_X:       return  "X";
  case CONSTR_CARRY:   return  "/";
  case CONSTR_NCARRY:  return  "N";
  case CONSTR_UCARRY:  return  "U";

  case CONSTR_NOTN:    return  "M";
  case CONSTR_NOTU:    return  "V";

  case CONSTR_I:       return  "I";
  case CONSTR_ICARRY:  return  "[";
  case CONSTR_II:      return  "]";
  case CONSTR_J:       return  "J";

  case CONSTR_Y:       return  "\\";
  case CONSTR_X3:      return  "3";
  case CONSTR_X5:      return  "5";
  case CONSTR_X7:      return  "7";
  case CONSTR_XA:      return  "A";
  case CONSTR_XB:      return  "B";
  case CONSTR_XC:      return  "C";
  case CONSTR_XD:      return  "D";
  case CONSTR_XE:      return  "E";

  case CONSTR_IMPOSSIBLE: return "#";

  default:;
    static char buf[11];
    snprintf (buf, sizeof(buf), "(%08x)", x);
    return buf;
  }
};

enum constraints to_const[256] = {
  ['-']  = CONSTR_EQ,
  ['x']  = CONSTR_NEQ,
  ['?']  = CONSTR_UNKOWN,
  ['=']  = CONSTR_SAME,
  ['<']  = CONSTR_XSAME,
  ['!']  = CONSTR_DIFF,
  ['>']  = CONSTR_XDIFF,
  ['0']  = CONSTR_ZERO,
  ['u']  = CONSTR_UP,
  ['1']  = CONSTR_ONE,
  ['n']  = CONSTR_DOWN,
  ['X']  = CONSTR_X,
  ['/']  = CONSTR_CARRY,
  ['N']  = CONSTR_NCARRY,
  ['U']  = CONSTR_UCARRY,

  ['M']  = CONSTR_NOTN,
  ['V']  = CONSTR_NOTU,

  ['I']  = CONSTR_I,
  ['[']  = CONSTR_ICARRY,
  [']']  = CONSTR_II,
  ['J']  = CONSTR_J,

  ['\\'] = CONSTR_Y,
  ['3']  = CONSTR_X3,
  ['5']  = CONSTR_X5,
  ['7']  = CONSTR_X7,
  ['A']  = CONSTR_XA,
  ['B']  = CONSTR_XB,
  ['C']  = CONSTR_XC,
  ['D']  = CONSTR_XD,
  ['E']  = CONSTR_XE,
  ['#']  = CONSTR_IMPOSSIBLE,
};


enum constraints constr_rotation_safe(enum constraints x) {
  uint32_t y = 0;
  for (uint32_t mask=0xff; mask; mask<<=8) {
    if (x & mask)
      y |= mask;
  }
  return y;
};

enum constraints constr_rotation_safe1(enum constraints x) {
  uint32_t y = 0;
  for (uint32_t mask=0xf; mask; mask<<=4) {
    if (x & mask)
      y |= mask;
  }
  return y;
};

enum constraints constr_rotation_safe2(enum constraints x) {
  uint32_t y = 0;
  for (uint32_t mask=0x3; mask; mask<<=2) {
    if (x & mask)
      y |= mask;
  }
  return y;
};

void print_reg(reg *r) {
  
  // Forget redundant information

  reg r2;
  memcpy(&r2, r, sizeof(reg));

#define idx(x) (((x)+WORD_SIZE)%WORD_SIZE)

  for (int i=WORD_SIZE-1; i>=0; i--) {
    
    state_bf x = 0;
    
    for (int j=0; j<8; j++) {
      if (r->bits[idx(i-1)] & (0xf << (4*j)))
        x |= 1<<j;
    }
    x *= 0x01010101;
    
    if ((~r->bits[i] & x & constr_rotation_safe(r->bits[i])) == 0)
      r2.bits[i] = constr_rotation_safe(r2.bits[i]);
    else if ((~r->bits[i] & x & constr_rotation_safe1(r->bits[i])) == 0)
      r2.bits[i] = constr_rotation_safe1(r2.bits[i]);
    else if ((~r->bits[i] & x & constr_rotation_safe2(r->bits[i])) == 0)
      r2.bits[i] = constr_rotation_safe2(r2.bits[i]);

    char *s;
    if ((s = constr_to_char(r2.bits[i]))) {
      printf("%s", s);
    } else {
      /*      int z = 4*__builtin_popcountl(r->bits[i] & x) / __builtin_popcountl(x);
      char s[5] = {'°', '¼', '½', '¾', '¹'};
      printf ("%c", s[z]);
      */
      printf ("@");
    }

  }

#undef idx
   
  /* for (int i=WORD_SIZE-1; i>=0; i--) { */
  /*   printf ("%s", constr_to_char(r2.bits[i])); */
  /* } */
}

void read_reg_from_str (reg *r, const char* s) {
  int i=0;
  for (; *s && i<WORD_SIZE; s++) {
    if (*s == '(') {
      r->bits[WORD_SIZE-1-i++] = strtoul(s+1, &s, 16);
      if (*s != ')') {
        printf ("Malformed string!\n");
        exit(-1);
      }
    } else {
      int c = to_const[*s];
      if (c == 0 && *s != '#') {
        printf ("Warning: Unknown char: `%c'\n", *s);
        c = CONSTR_UNKOWN;
      }
      r->bits[WORD_SIZE-1-i++] = c;
    }
  }

  if (*s)
    printf ("Warning: string to long\n");

  if (i < WORD_SIZE) {
    printf ("Warning: string to short\n");
    for (; i<WORD_SIZE; i++)
      r->bits[i] = CONSTR_UNKOWN;
  }
}

//#define SPARSE_SIZE 16
//#define SPARSE_PAR 3

int eq_count_upto(equation *eq) {

  sparse_automaton_t *at = eq->transition_sparse;
  if (at->data == NULL) {
    fprintf (stderr, "Error: transition automaton not defined.\n");
    return -1;
  }

  int npar = at->nb_params;
  int nst  = at->nb_states;
  int SPARSE_SIZE = at->sparse_size;
  parameter *in = eq->p;

  state_bf (*aa)[SPARSE_SIZE][npar+1] = (void*) at->data;

  struct reg par0[npar];
  for (int k=0; k<npar; k++) {
    memcpy(&par0[k], in[k].r, sizeof(struct reg));
    par0[k].bits[(0+in[k].rot) % WORD_SIZE] =
      constr_rotation_safe  (par0[k].bits[(0+in[k].rot) % WORD_SIZE]);
    par0[k].bits[(1+in[k].rot) % WORD_SIZE] =
      constr_rotation_safe2 (par0[k].bits[(1+in[k].rot) % WORD_SIZE]);
  }

  int p[WORD_SIZE+1][nst];
  memset (p, 0, sizeof(p));
  p[0][0] = 1;
  for (int i=0; i<WORD_SIZE; i++) {
    int ok=0;
    for (int j=0; j<nst; j++) {
      if (p[i][j]) {
        for (int x=0; x<SPARSE_SIZE; x++) {
          int k;
          for (k=0; k<npar; k++) {
            if (!(par0[k].bits[(i+in[k].rot) % WORD_SIZE] & aa[j][x][k]))
              break;
          }
          if (k == npar) {
            p[i+1][aa[j][x][npar]] = 1;
            ok = 1;
          }
        }
      }
    }
    if (!ok)
      return i;
  }

  return WORD_SIZE;
}

int eq_compatible(equation *eq) {
  return eq_count_upto(eq) == WORD_SIZE;
}

double eq_count(equation *eq) {

  sparse_automaton_t *at = eq->transition_sparse;
  if (at->data == NULL) {
    fprintf (stderr, "Error: transition automaton not defined.\n");
    return -1;
  }

  int npar = at->nb_params;
  int nst  = at->nb_states;
  int sparse_size = at->sparse_size;
  parameter *in = eq->p;

  state_bf (*aa)[sparse_size][npar+1] = (void*) at->data;

  struct reg par0[npar];
  for (int k=0; k<npar; k++) {
    memcpy(&par0[k], in[k].r, sizeof(struct reg));
    par0[k].bits[(0+in[k].rot) % WORD_SIZE] =
      constr_rotation_safe  (par0[k].bits[(0+in[k].rot) % WORD_SIZE]);
    par0[k].bits[(1+in[k].rot) % WORD_SIZE] =
      constr_rotation_safe2 (par0[k].bits[(1+in[k].rot) % WORD_SIZE]);
  }

  bignum p[WORD_SIZE+1][nst];
  memset (p, 0, sizeof(p));
  p[0][0] = 1;
  for (int i=0; i<WORD_SIZE; i++) {
    int ok=0;
    for (int j=0; j<nst; j++) {
      if (p[i][j]) {
        for (int x=0; x<sparse_size; x++) {
          int k;
          for (k=0; k<npar; k++) {
            if (!(par0[k].bits[(i+in[k].rot) % WORD_SIZE] & aa[j][x][k]))
              break;
          }
          if (k == npar) {
            p[i+1][aa[j][x][npar]] += p[i][j];
            ok = 1;
          }
        }
      }
      //      printf ("/%f", (double) p[i][j]);
    }
    //    printf ("\n");
    if (!ok)
      return 0;
  }

  bignum ret = 0;
  for (int i=0; i<nst; i++)
    ret += p[WORD_SIZE][i];

  //  print_reg(in[0].r);
  //  printf (": %f\n\n", (double) ret);

  return ret;
}

int eq_count_propagate(equation *eq) {

  sparse_automaton_t *at = eq->transition_sparse;
  if (at->data == NULL) {
    fprintf (stderr, "Error: transition automaton not defined.\n");
    return -1;
  }

  int npar = at->nb_params;
  int nst  = at->nb_states;
  int SPARSE_SIZE = at->sparse_size;
  parameter *in = eq->p;

  state_bf (*aa)[SPARSE_SIZE][npar+1] = (void*) at->data;

  struct reg par0[npar];
  for (int k=0; k<npar; k++) {
    memcpy(&par0[k], in[k].r, sizeof(struct reg));
    par0[k].bits[(0+in[k].rot) % WORD_SIZE] =
      constr_rotation_safe  (par0[k].bits[(0+in[k].rot) % WORD_SIZE]);
    par0[k].bits[(1+in[k].rot) % WORD_SIZE] =
      constr_rotation_safe2 (par0[k].bits[(1+in[k].rot) % WORD_SIZE]);
  }
  int p[WORD_SIZE+1][nst];
  memset (p, 0, sizeof(p));
  p[0][0] = 1;
  for (int i=0; i<WORD_SIZE; i++) {
    int ok=0;
    for (int j=0; j<nst; j++) {
      if (p[i][j]) {
        for (int x=0; x<SPARSE_SIZE; x++) {
          int k;
          for (k=0; k<npar; k++) {
            if (!(par0[k].bits[(i+in[k].rot) % WORD_SIZE] & aa[j][x][k]))
              break;
          }
          if (k == npar) {
            p[i+1][aa[j][x][npar]] = 1;
            ok = 1;
          }
        }
      }
    }
    if (!ok)
      return -1;
  }

  for (int j=0; j<nst; j++) {
    if (p[WORD_SIZE][j]==1)
      p[WORD_SIZE][j]=2;
  }

  struct reg par2[npar];
  memset(par2, 0, sizeof(par2));

  for (int i=WORD_SIZE-1; i>=0; i--) {
    for (int j=0; j<nst; j++) {
      if (p[i][j]) {
        for (int x=0; x<SPARSE_SIZE; x++) {
          int k;
          for (k=0; k<npar; k++) {
            if (!(par0[k].bits[(i+in[k].rot) % WORD_SIZE] & aa[j][x][k]))
              break;
          }
          if (k == npar && p[i+1][aa[j][x][npar]] == 2) {
            p[i][j] = 2;
            for (k=0; k<npar; k++) {
              par2[k].bits[(i+in[k].rot) % WORD_SIZE] |= aa[j][x][k];
            }
          }
        }
      }
    }
  }

  int new = 0;

  int newr[npar];
  memset(newr, 0, sizeof(newr));

  for (int k=0; k<npar; k++) {
    par2[k].bits[(0+in[k].rot) % WORD_SIZE] =
      constr_rotation_safe  (par2[k].bits[(0+in[k].rot) % WORD_SIZE]);
    par2[k].bits[(1+in[k].rot) % WORD_SIZE] =
      constr_rotation_safe2 (par2[k].bits[(1+in[k].rot) % WORD_SIZE]);

    for (int i=0; i<WORD_SIZE; i++) {
      state_bf n = in[k].r->bits[i] & par2[k].bits[i];
      if (n != in[k].r->bits[i]) {
        in[k].r->bits[i] = n;
        //        printf ("Sign constraint: %i.%i:%c\n", k, i, constr_to_char[n]);
        new ++;
        newr[k]++;
      }
    }
  }
  
  assert(eq_compatible(eq));

  /*
  printf ("Quick propagate:\n");
  for (int i=0; i<npar; i++) {
    print_reg(in[i].r);
    printf ("\n");
  }
  */

  return new;
}

size_t mmap_at (char *file, void **at) {
  int fd = open(file, O_RDONLY);
  
  if (fd < 0) {
    char buffer[32];
    snprintf (buffer, sizeof(buffer), "open %s", file);
    perror (buffer);
    return -1;
  }
  
  struct stat st;
  fstat(fd, &st);
  size_t size = st.st_size;
  *at = mmap(NULL, size,
             PROT_READ, MAP_SHARED, fd, 0);
  if (*at == MAP_FAILED) {
    perror ("mmap:");
    return -1;
  }
  return size;
}

int main (int argc, char* argv[]) {
  int propagate = 0;
  int cost = 0;

  int opt;

  int nvar = -1;
  int npar = -1;
  int xor = 0;
  equation e;
  memset(&e, 0, sizeof(e));

  while ((opt = getopt(argc, argv, "w:vpcn:m:s:")) != -1) {

    switch (opt) {
    case 'v':
      propagate = 0;
      cost = 0;
      break;
    case 'p':
      propagate = 1;
      break;
    case 'c':
      cost = 1;
      break;
    case 'm':
      npar = atoi(optarg);
      break;
    case 'w':
      WORD_SIZE = atoi(optarg);
      break;
    case 's':
      if (mmap_at(optarg, &e.transition_sparse) == -1)
        return -1;
      break;

    default:
      printf ("Unknown option: `%c'\n", opt);
    case '?':
    usage:
      printf ("Usage: %s [-w n] [-v|-p|-c] [-d file] [-t file] [-n n] [-m n]\n"
              "       -- cst1 [cst2 ...]\n"
              "\n"
              "    -w <n>: set word size\n"
              "\n"
              "    -v: only verify constraints (default) [requires -d]\n"
              "    -p: propagate constraints             [requires -d]\n"
              "    -c: count solutions                   [requires -t]\n"
              "\n"
              "    -s <file>: sparse automaton\n"
              "\n"
              "    -n <n>: number of variables\n"
              "    -m <n>: number of parameters\n", argv[0]);
    return -1;
    }
  }

  // Sanity checks 

  if (argc == optind) {
    fprintf(stderr, "Arguments needed\n");
    goto usage;
  }

  if (e.transition_sparse == NULL) {
    fprintf(stderr, "Automaton needed (use -s option)\n");
    goto usage;
  }

  npar = e.transition_sparse->nb_params;

  if (npar != -1) {
    if (npar != argc - optind) {
      fprintf(stderr, "Wrong number of parameters (%i v. %i)\n",
              npar, argc - optind);
      goto usage;
    }
  } else {
    npar = argc - optind;
  }

  if (npar > MAX_PARAMS) {
    fprintf(stderr, "To many paramters!\nPlease modify MAX_PARAMS and recompile\n");
    exit (-2);
  }

  reg regs[MAX_PARAMS];

  for (int i=0; i<argc-optind; i++) {
    e.p[i].r = &regs[i];
    e.p[i].rot = 0;
    read_reg_from_str(&regs[i], argv[optind+i]);
  }

  /* print_reg(&regs[0]); printf (" "); */
  /* print_reg(&regs[1]); printf (" "); */
  /* print_reg(&regs[2]); printf ("\n"); */

  if (!eq_compatible(&e)) {
    printf ("System fails at %2i\n", eq_count_upto(&e));
  } else {
    printf ("System is compatible!\n");
    if (propagate) {
      printf ("Propagate:\n");

      // First, normalize the constraints
      for (int i=0; i<npar; i++) {
	equation ee = {
	  { e.p[i] },
	  &xx_count,
	};
	eq_count_propagate(&ee);
      }

      int x = eq_count_propagate(&e);
      printf ("%i new constraints\nNew parameters:\n", x);
      for (int i=0; i<npar; i++) {
        print_reg(&regs[i]);
        printf ("\n");
      }
    }

    if (cost) {
      bignum p0 = eq_count(&e);
      printf ("#Sol: %llu (2^%.1f)\n", (unsigned long long)p0, log((double) p0)/log(2));

      if (npar > 1) {

        // Count the number of possible values for each parameter
        bignum counts[npar];
        bignum prod = 1;
        for (int i=0; i<npar; i++) {
          equation ee = {
            { e.p[i] },
            &xx_count,
          };
          counts[i] = eq_count(&ee);
          prod *= counts[i];
        }
        

        // Now compute probabilities with a random choice a a given parameter
        
        for (int i=0; i<npar; i++) {
          printf ("Cost[%i]: %.1f\n", i,
                  -log((double)(p0*counts[i]/prod))/log(2));
        }
      }
    }

  }

  return 0;
}
