/* Primitive parser */
%{
  open Exp

  let varstbl = Hashtbl.create 8

  let makevar name exp = 
    Hashtbl.add varstbl name exp 

  let callvar name =
    Hashtbl.find varstbl name

  let macrostbl = Hashtbl.create 8

  let makemacro name exp = 
    Hashtbl.add macrostbl name exp 

  let callmacro name l =
    let rec subst = function
      | Sum  (i, x, y)    -> Sum  (i, subst x, subst y)
      | Sum3 (i, x, y, z) -> Sum3 (i, subst x, subst y, subst z)
      | Mux (i, l)        -> Mux (i, List.map subst l)
      | HW  (i, w, x)     -> HW  (i, w, subst x)
      | Bool (b, x, y)        -> Bool (b, subst x, subst y)
      | Var i when i < 0  -> List.nth l (-1-i)
      | Var i             -> Var i
      | Param i           -> Param i
      | Const i           -> Const i
    in subst (Hashtbl.find macrostbl name)

  (* Note: during parsing, the sum number is actually the initial carry *)

  let makesyst e =
    let nbsum = ref 0 in
    let nbvar = ref 0 in
    let nbpar = ref 0 in
    let htbl = Hashtbl.create 8 in
    let htbl2 = Hashtbl.create 8 in
    let initcarry:int list ref = ref [] in
    let rec f = function
      | Sum  (c, x, y)    ->
          let i = try
            Hashtbl.find htbl (c,x,y,Const false)
          with Not_found ->
            let n = !nbsum in
            Hashtbl.add htbl (c,x,y,Const false) n;
            initcarry := c::!initcarry;
            incr nbsum;
            n;
          in Sum  (i, f x, f y);
      | Sum3 (c, x, y, z)    ->
          let i = try
            Hashtbl.find htbl (c,x,y,z)
          with Not_found ->
            let n = !nbsum in
            Hashtbl.add htbl (c,x,y,z) n;
            initcarry := c::!initcarry;
            incr nbsum;
            n;
          in Sum3  (i, f x, f y, f z);
      | HW  (_, w, x)     ->
        let i = try
            Hashtbl.find htbl2 x
          with Not_found ->
            let n = !nbsum in
            Hashtbl.add htbl2 x n;
            initcarry := 0::!initcarry;
            incr nbsum;
            n;
        in HW  (i, w, f x)
      | Mux (i, l)        -> Mux (i, List.map f l)
      | Bool (b, x, y)    -> Bool (b, f x, f y)
      | Var i             -> if i+1 > !nbvar then nbvar := i+1; Var i
      | Param i           -> if i+1 > !nbpar then nbpar := i+1; Param i
      | Const i           -> Const i
    in let ee = List.map f e in {
      exps      = ee;
      nbvar     = !nbvar;
      nbparam   = !nbpar;
      nbsum     = !nbsum;
      initcarry = Array.of_list (List.rev !initcarry);
    }
%}

%token <int>  Tvar
%token <int>  Tpar
%token <int>  Tmacro
%token <int>  Ttmp
%token <int>  Tmpar
%token <int>  Tmux
%token <int>  Thw
%token <bool> Tcst
%token Tminus
%token Tplus
%token Txor
%token Tor
%token Tand
%token Txor2
%token Tor2
%token Tand2
%token Topen
%token Tclose
%token Tcomma
%token Tsep
%token Teq
%token Teq2
%token Tnot
%token Tdef
%token Tdefvar
%token EOF

%left Txor2 Tor2 Tand2 Teq2
%left Txor Tor Tand Teq Tplus Tminus

%start system
%type <Exp.syst> system
%type <Exp.exp list> equations
%type <Exp.exp> equation

%%

system:
  | equations { makesyst $1 }

equations:
| EOF { [] }
| equation EOF { [$1] }
| equation; Tsep; equations { $1::$3 }
;

equation:
| expression { $1 }
| macro { Const true }
| Thw Topen expression Tclose { HW(0,$1,$3) }
| Ttmp Tdefvar expression { makevar $1 $3; Const true }

expression:
| Tcst { Const $1 }
| Tvar { Var $1 }
| Tpar { Param $1 }
| Tmpar { Var (-1-$1) }
| Tnot expression { Bool ((fun x y -> not x), $2, Const true) }
| Ttmp { callvar $1 }
| expression Txor expression  { Bool ((<>), $1, $3) }
| expression Teq  expression  { Bool ((=) , $1, $3) }
| expression Tor  expression  { Bool ((||), $1, $3) }
| expression Tand expression  { Bool ((&&), $1, $3) }
| expression Txor2 expression { Bool ((<>), $1, $3) }
| expression Teq2  expression { Bool ((=) , $1, $3) }
| expression Tor2  expression { Bool ((||), $1, $3) }
| expression Tand2 expression { Bool ((&&), $1, $3) }
| expression Tplus expression  {
  match $1 with
    | Sum(c,a,b) -> output_string stderr "Sum3!\n"; Sum3(c,a,b,$3)
    | _          -> Sum (0, $1, $3)
}
| expression Tminus expression {
  match $1 with
    | Sum(c,a,b) -> output_string stderr "Sum3!\n"; Sum3(c+1,a,b,Bool((<>), $3, Const true))
    | _          -> Sum (1, $1, Bool((<>), $3, Const true))
}
| Topen expression Tclose { $2 }
| Tmacro Topen expressions Tclose { callmacro $1 $3 }
| Tmux   Topen expressions Tclose {
    if (List.length $3) <> ($1 + (1 lsl $1)) then
      failwith "Bad number of arguments for Mux call";
    Mux ($1, $3) }

expressions:
| expression { [$1] }
| expression Tcomma expressions { $1::$3 }

macro:
  | Tmacro Tdef expression { makemacro $1 $3 }
